#!/usr/bin/env python3
import graphsurgeon as gs
import numpy as np
import tensorflow as tf
import tensorrt as trt
import uff

try:
    import common
except ImportError:
    print('Need to import /usr/src/tensorrt/samples/python/common.py')
    print('e.g. export PYTHONPATH=$PYTHONPATH:/usr/src/tensorrt/samples/python/')
    raise

# https://devtalk.nvidia.com/default/topic/1038494/tensorrt/logicerror-explicit_context_dependent-failed-invalid-device-context-no-currently-active-context-/post/5284290/#5284290
import pycuda.autoinit
import pycuda.driver as cuda
from common import HostDeviceMem

num_anchors = 1
num_classes = 2

loc_data = tf.placeholder(tf.float32, [num_anchors * 4, 1, 1], name='loc_data')
conf_data = tf.placeholder(tf.float32, [num_anchors * num_classes, 1, 1], name='conf_data')
priorbox_data = tf.placeholder(tf.float32, [2, num_anchors * 4, 1], name='priorbox_data')

NMS = gs.create_plugin_node(
    name="NMS",
    op="NMS_TRT",
    shareLocation=1,
    varianceEncodedInTarget=0,
    backgroundLabelId=0,
    confidenceThreshold=1e-8,
    nmsThreshold=0.6,
    topK=2,
    keepTopK=2,
    numClasses=num_classes,
    inputOrder=[0, 1, 2],  # loc_data, conf_data, priorbox_data
    confSigmoid=0,
    isNormalized=0,
    codeType=0,  # CORNER = 0
    )

NMS.input.extend([tensor.op.name for tensor in [loc_data, conf_data, priorbox_data]])

dynamic_graph = gs.DynamicGraph(tf.get_default_graph().as_graph_def())
dynamic_graph.append(NMS)

serialized_uff = uff.from_tensorflow(
    dynamic_graph.as_graph_def(),
    output_nodes=['NMS'],
    output_filename='/tmp/trt_NMS_test.uff',
    text=False,
    )
def allocate_buffers(engine):
    """Allocates host and device buffer for TRT engine inference.

    This function is similair to the one in ../../common.py, but
    converts network outputs (which may be np.int32) appropriately
    before writing them to Python buffer. This is needed, since
    TensorRT plugins doesn't support output type description, and
    in our particular case, we use NMS plugin as network output.

    Args:
        engine (trt.ICudaEngine): TensorRT engine

    Returns:
        inputs [HostDeviceMem]: engine input memory
        outputs [HostDeviceMem]: engine output memory
        bindings [int]: buffer to device bindings
        stream (cuda.Stream): cuda stream for engine inference synchronization
    """
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()

    # Current NMS implementation in TRT only supports DataType.FLOAT but
    # it may change in the future, which could brake this sample here
    # when using lower precision [e.g. NMS output would not be np.float32
    # anymore, even though this is assumed in binding_to_type]
    binding_to_type = {"loc_data": np.float32, "conf_data":np.float32,
            "priorbox_data":np.float32, "NMS": np.float32, "NMS_1": np.int32}

    for binding in engine:
        print(binding)
        size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
        dtype = binding_to_type[str(binding)]
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream

TRT_LOGGER = trt.Logger(trt.Logger.Severity.INFO)
with trt.Builder(TRT_LOGGER) as builder:
    builder.max_workspace_size = common.GiB(1)
    trt.init_libnvinfer_plugins(TRT_LOGGER, '')
    with builder.create_network() as network:
        uff_parser = trt.UffParser()
        uff_parser.register_input('loc_data', [num_anchors * 4, 1, 1])
        uff_parser.register_input('conf_data', [num_anchors * num_classes, 1, 1])
        uff_parser.register_input('priorbox_data', [2, num_anchors * 4, 1])
        uff_parser.register_output('NMS')
        uff_parser.parse('/tmp/trt_NMS_test.uff', network)

        with builder.build_cuda_engine(network) as engine:
            inputs, outputs, bindings, stream = allocate_buffers(engine)
            print(engine.num_bindings)
            with engine.create_execution_context() as context:
                loc_data_np = np.array([0, 0, 0, 0], dtype=np.float32)
                conf_data_np = np.array([0, 1], dtype=np.float32)
                priorbox_data_np = np.array([.1, .1, .50, .50, 0, 0, 0, 0], dtype=np.float32)

                np.copyto(inputs[0].host, loc_data_np.ravel())
                np.copyto(inputs[1].host, conf_data_np.ravel())
                np.copyto(inputs[2].host, priorbox_data_np.ravel())

                results = common.do_inference(context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream)

                detection_output = results[0].reshape((-1, 7))
                print(detection_output)
                print(results)
                num_detections = int(results[1])
                print('Number of detections: {}'.format(num_detections))
                print(detection_output[:num_detections, :])
